@@ -655,6 +655,8 @@ static void ext2fs_journal_release(ext2_filsys fs, journal_t *journal,
 		if (fs->io != fs->journal_io)
 			io_channel_close(fs->journal_io);
 		fs->journal_io = NULL;
+		free(fs->journal_name);
+		fs->journal_name = NULL;
 	}
 
 #ifndef USE_INODE_IO
@@ -791,14 +793,14 @@ errcode_t ext2fs_run_ext3_journal(ext2_filsys *fsp)
 		kbytes_written = stats->bytes_written >> 10;
 
 	ext2fs_mmp_stop(fs);
-	fsname = strdup(fs->device_name);
+	fsname = fs->device_name;
+	fs->device_name = NULL;
 	fsflags = fs->flags;
 	fsblocksize = fs->blocksize;
 	ext2fs_free(fs);
-	retval = ext2fs_open(fsname, fsflags,
-			     0, fsblocksize, io_ptr,
-			     fsp);
-	free(fsname);
+	*fsp = NULL;
+	retval = ext2fs_open(fsname, fsflags, 0, fsblocksize, io_ptr, fsp);
+	ext2fs_free_mem(&fsname);
 	if (retval)
 		return retval;
 
